/*
 * GPJCamera.java v1.4
 *
 * Copyright (C) 2010  by Giuseppe Perniola
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

public class GPJCamera 
{

    public static final int TYPE_MOVE_FORWARD   = 1;
    public static final int TYPE_MOVE_BACK      = 2;
    public static final int TYPE_MOVE_LEFT      = 4;
    public static final int TYPE_MOVE_RIGHT     = 8;

    public GPJVector3D      m_pos;
    public GPJVector3D      m_angles;
    
    protected int           m_screenWidth;
    protected int           m_screenHeight;
    protected int           m_screenCenterX;
    protected int           m_screenCenterY;
    protected int[]         m_screenBuffer;
    protected int[]         m_zBuffer;
    protected float         m_screenDistX;
    protected float         m_screenDistY;
    protected float         m_nearZ;
    protected float         m_farZ;
    protected GPJMatrix     m_matrixRot;
    protected GPJMatrix     m_matrixMove;
    
    private float           m_zBufferA;
    private float           m_zBufferB;
    private GPJVector3D     m_vectorTemp1;
    private GPJVector3D     m_vectorTemp2;
    
    public GPJCamera(float posX,
                     float posY,
                     float posZ,
                     float angleX,
                     float angleY,
                     float angleZ)
    {
        m_pos           = new GPJVector3D();
        m_angles        = new GPJVector3D();

        m_matrixRot     = new GPJMatrix();
        m_matrixMove    = new GPJMatrix();
        
        m_vectorTemp1   = new GPJPoint3D();
        m_vectorTemp2   = new GPJPoint3D();

        m_pos.set(posX,
                posY,
                posZ);

        m_angles.set(angleX,
                angleY,
                angleZ);
        
        updateAngulation();

        m_screenBuffer  = null;
        m_zBuffer       = null;
    }
    
    public boolean setScreen(int    fov,
                             int    screenWidth,
                             int    screenHeight,
                             float  nearZ,
                             float  farZ)
    {
        float angle = (float)Math.PI / 180.0f * (float)fov;

        if (fov <= 0                                        ||
                fov >= 180                                  ||
                screenWidth <= 2                            ||
                screenHeight <= 2                           ||
                Math.abs(nearZ) <= GPJEngine3D.TOLLERANCE   ||
                farZ - m_nearZ <= GPJEngine3D.TOLLERANCE)
            return false;
        
        m_screenWidth   = screenWidth;
        m_screenHeight  = screenHeight;
        m_screenCenterX = screenWidth >> 1;
        m_screenCenterY = screenHeight >> 1;
        m_screenDistX   = (float)screenWidth / 2.0f * (float)Math.tan(angle);
        m_screenDistY   = (float)screenHeight / 2.0f * (float)Math.tan(angle);
        m_nearZ         = nearZ;
        m_farZ          = farZ;

        m_zBufferA = m_farZ / (m_farZ - m_nearZ);
        m_zBufferB = m_farZ * m_nearZ / (m_nearZ - m_farZ);

        m_screenBuffer  = new int[screenWidth * screenHeight];
        m_zBuffer       = new int[screenWidth * screenHeight];

        return true;
    }
       
    public void move(int    typeMove,
                     float  value)
    {
        m_vectorTemp1.m_x = 0.0f;
        m_vectorTemp1.m_y = 0.0f;
        m_vectorTemp1.m_z = 0.0f;
        if (typeMove == TYPE_MOVE_FORWARD)
            m_vectorTemp1.m_z = value;
        else if (typeMove == TYPE_MOVE_BACK)
            m_vectorTemp1.m_z = -value;
        else if (typeMove == TYPE_MOVE_LEFT)
            m_vectorTemp1.m_x = -value;
        else if (typeMove == TYPE_MOVE_RIGHT)
            m_vectorTemp1.m_x = value;

        GPJVector3D.multiplyVectorMatrix(m_vectorTemp1,
                m_matrixMove,
                m_vectorTemp2);
        GPJVector3D.addVectors(m_pos,
                m_vectorTemp2,
                m_pos);
    }
    
    public void rotate(float diffAngleX,
                       float diffAngleY,
                       float diffAngleZ)
    {
        float angleX;
        float angleY;
        float angleZ;
        
        angleX = m_angles.m_x + diffAngleX;
        angleY = m_angles.m_y + diffAngleY;
        angleZ = m_angles.m_z + diffAngleZ;
        m_angles.set(angleX,
                angleY,
                angleZ);
        updateAngulation();
    }
    
    public void updateAngulation()
    {
        m_matrixRot.setIdentity();
        m_matrixRot.rotate(m_angles.m_x,
                m_angles.m_y,
                m_angles.m_z);
        GPJMatrix.invertMatrix(m_matrixRot,
                m_matrixMove);
    }

    public void clearZBuffer()
    {
        int i;

        if (m_zBuffer == null)
            return;

        for (i = 0; i < m_zBuffer.length; i ++)
            setValueZBuffer(i,
                    m_farZ,
                    false);
    }

    public boolean setValueZBuffer(int      offset,
                                   float    z,
                                   boolean  check)
    {
        int value;

        if (m_zBuffer == null   ||
                z == 0.0f       ||
                offset < 0      ||
                offset >= m_zBuffer.length)
            return false;

        value = (int)((float)(1 << 31) * (m_zBufferA + m_zBufferB / z));

        if (check &&
                value < m_zBuffer[offset])
            return false;

        m_zBuffer[offset] = value;

        return true;
    }
    
}
