/*
 * GPJEngine3D.java v1.5
 *
 * Copyright (C) 2010  by Giuseppe Perniola
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.*;

public class GPJEngine3D 
{

    public static final int         FLAT_SHADING    = 0;
    public static final int         GOURAUD_SHADING = 1;

    public int                      m_shading;

    protected static final float    TOLLERANCE      = 0.0001f;

    private int                     m_numVertexesClipped;
    private Vector                  m_listVertexesOriginal;
    private Vector                  m_listVertexesProcessed;
    private Vector                  m_listPointsOriginal;
    private Vector                  m_listPointsProcessed;
    private Vector                  m_listLines;
    private Vector                  m_listTriangles;
    private GPJCamera               m_camera;
    private GPJPoint3D              m_vertexTemp;
    private GPJPoint3D[]            m_listVertexesClipped   = null;
    private GPJColor                m_colorTemp;
    private GPJColor                m_colorTemp2;
    private GPJColor                m_colorInterpolated;
    private GPJColor                m_colorBackground;
    private GPJVector2D             m_uvInterpolated;
    private GPJVector3D             m_renderOldP;
    private GPJVector3D             m_renderOldM;
    private GPJVector3D             m_renderOldN;
    private GPJVector3D             m_renderNewP;
    private GPJVector3D             m_renderNewM;
    private GPJVector3D             m_renderNewN;
    private GPJVector3D             m_renderA;
    private GPJVector3D             m_renderB;
    private GPJVector3D             m_renderC;
    private GPJVector3D             m_renderD;
    private GPJVector3D             m_vectorTemp;
    private GPJLight                m_lightAmbient          = null;
    private GPJLight                m_lightOther            = null;
    
    public GPJEngine3D(GPJColor colorBackground,
                       int      shading)
    {
        int i;

        m_shading = shading;

        m_listVertexesOriginal  = new Vector();
        m_listVertexesProcessed = new Vector();
        m_listPointsOriginal    = new Vector();
        m_listPointsProcessed   = new Vector();
        m_listLines             = new Vector();
        m_listTriangles         = new Vector();
        
        m_vertexTemp = new GPJPoint3D();

        m_colorTemp         = new GPJColor();
        m_colorTemp2        = new GPJColor();
        m_colorInterpolated = new GPJColor();
        m_colorBackground   = colorBackground;
        
        m_uvInterpolated = new GPJVector2D();

        m_renderOldP    = new GPJVector3D();
        m_renderOldM    = new GPJVector3D();
        m_renderOldN    = new GPJVector3D();
        m_renderNewP    = new GPJVector3D();
        m_renderNewM    = new GPJVector3D();
        m_renderNewN    = new GPJVector3D();
        m_renderA       = new GPJVector3D();
        m_renderB       = new GPJVector3D();
        m_renderC       = new GPJVector3D();
        m_renderD       = new GPJVector3D();
        m_vectorTemp    = new GPJVector3D();

        m_lightAmbient  = null;
        m_lightOther    = null;
    }

    public boolean addVertex(GPJPoint3D vertex)
    {
        if (vertex != null)
        {
            m_listVertexesOriginal.addElement(vertex);
            m_listVertexesProcessed.addElement(new GPJPoint3D());
            
            return true;
        }
        
        return false;
    }

    public boolean removeVertex(GPJPoint3D vertex)
    {
        if (vertex != null)
        {
            if (m_listVertexesOriginal.removeElement(vertex))
            {
                m_listVertexesProcessed.removeElementAt(0);

                return true;
            }
            else
                return false;
        }

        return false;
    }

    public boolean addPoint(GPJPoint3D point)
    {
        if (point != null)
        {
            m_listPointsOriginal.addElement(point);
            m_listPointsProcessed.addElement(new GPJPoint3D());

            return true;
        }

        return false;
    }

    public boolean removePoint(GPJPoint3D point)
    {
        if (point != null)
        {
            if (m_listPointsOriginal.removeElement(point))
            {
                m_listPointsProcessed.removeElementAt(0);

                return true;
            }
            else
                return false;
        }

        return false;
    }

    public boolean addLine(GPJLine line)
    {
        int i;

        if (line != null)
        {
            m_listLines.addElement(line);
            for (i = 0; i < GPJLine.NUM_VERTEXES; i ++)
                line.m_listVertexesProcessed[i] = (GPJPoint3D)m_listVertexesProcessed.elementAt(line.m_listNumVertexes[i]);

            return true;
        }

        return false;
    }

    public boolean removeLine(GPJLine line)
    {
        if (line != null)
            return m_listLines.removeElement(line);

        return false;
    }
    
    public boolean addTriangle(GPJTriangle triangle)
    {
        int i;

        if (triangle != null)
        {
            m_listTriangles.addElement(triangle);
            for (i = 0; i < GPJTriangle.NUM_VERTEXES; i ++)
                triangle.m_listVertexesProcessed[i] = (GPJPoint3D)m_listVertexesProcessed.elementAt(triangle.m_listNumVertexes[i]);
            
            return true;
        }
        
        return false;
    }

    public boolean removeTriangle(GPJTriangle triangle)
    {
        if (triangle != null)
            return m_listTriangles.removeElement(triangle);

        return false;
    }
 
    public void addLight(GPJLight light)
    {
        if (light != null)
        {
            if (light.m_type == GPJLight.TYPE_AMBIENT)
                m_lightAmbient = light;
            else
                m_lightOther = light;
        }
        else
        {
            m_lightOther = null;
        }
    }
    
    public boolean setCamera(GPJCamera camera)
    {
        if (camera != null)
        {
            m_camera = camera;
            
            return true;
        }
        
        return false;
    }

    public void init(boolean clearZBuffer,
                     boolean clearBackground)
    {
        int i;

        if (clearZBuffer)
            m_camera.clearZBuffer();

        if (clearBackground)
            for (i = 0; i < m_camera.m_screenBuffer.length; i ++)
                m_camera.m_screenBuffer[i] = m_colorBackground.m_RGB;

        m_listVertexesOriginal.clear();
        m_listVertexesProcessed.clear();
        m_listPointsOriginal.clear();
        m_listPointsProcessed.clear();
        m_listLines.clear();
        m_listTriangles.clear();
    }
    
    public void render()
    {
        int         i;
        Enumeration listVertexesOriginal;
        Enumeration listVertexesProcessed;
        Enumeration listPointsOriginal;
        Enumeration listPointsProcessed;
        GPJPoint3D  vertexOriginal;
        GPJPoint3D  vertexProcessed;
        GPJPoint3D  pointOriginal;
        GPJPoint3D  pointProcessed;

        if (m_listVertexesClipped == null ||
                m_listVertexesClipped.length < m_listVertexesOriginal.size() * 2)
        {
            m_listVertexesClipped = new GPJPoint3D[m_listVertexesOriginal.size() * 2];
            for (i = 0; i < m_listVertexesClipped.length; i ++)
                m_listVertexesClipped[i] = new GPJPoint3D();

            System.gc();
        }
        else
            for (i = 0; i < m_listVertexesClipped.length; i ++)
                m_listVertexesClipped[i].reset();

        calcNormalsAndGouraudShading();

        for (listVertexesOriginal = m_listVertexesOriginal.elements(),
             listVertexesProcessed = m_listVertexesProcessed.elements(); listVertexesOriginal.hasMoreElements();)
        {
            vertexOriginal  = (GPJPoint3D)listVertexesOriginal.nextElement();
            vertexProcessed = (GPJPoint3D)listVertexesProcessed.nextElement();

            vertexProcessed.reset();
            
            m_vertexTemp.set(vertexOriginal.m_x - m_camera.m_pos.m_x,
                    vertexOriginal.m_y - m_camera.m_pos.m_y,
                    vertexOriginal.m_z - m_camera.m_pos.m_z);
            GPJVector3D.multiplyVectorMatrix(m_vertexTemp,
                    m_camera.m_matrixRot,
                    vertexProcessed);

            GPJColor.copyColor(vertexOriginal.m_colorProcessed,
                    vertexProcessed.m_color);

            GPJVector2D.copyVector(vertexOriginal.m_uv,
                    vertexProcessed.m_uv);

            if (vertexProcessed.m_z >= m_camera.m_nearZ &&
                    vertexProcessed.m_z <= m_camera.m_farZ)
                vertexProcessed.m_trasformTo2D = true;
        }

        for (listPointsOriginal = m_listPointsOriginal.elements(),
             listPointsProcessed = m_listPointsProcessed.elements(); listPointsOriginal.hasMoreElements();)
        {
            pointOriginal   = (GPJPoint3D)listPointsOriginal.nextElement();
            pointProcessed  = (GPJPoint3D)listPointsProcessed.nextElement();

            pointProcessed.reset();

            m_vertexTemp.set(pointOriginal.m_x - m_camera.m_pos.m_x,
                    pointOriginal.m_y - m_camera.m_pos.m_y,
                    pointOriginal.m_z - m_camera.m_pos.m_z);
            GPJVector3D.multiplyVectorMatrix(m_vertexTemp,
                    m_camera.m_matrixRot,
                    pointProcessed);

            GPJColor.copyColor(pointOriginal.m_color,
                    pointProcessed.m_color);

            if (pointProcessed.m_z >= m_camera.m_nearZ &&
                    pointProcessed.m_z <= m_camera.m_farZ)
                pointProcessed.m_trasformTo2D = true;
        }

        clipLines();

        clipTriangles();

        trasformVertexes3Dto2D();

        trasformPoints3Dto2D();

        draw();
    }

    private void calcNormalsAndGouraudShading()
    {
        int         a;
        int         index1;
        int         numNormals;
        float       ratioColorLightOther;
        Enumeration listTriangles;
        Enumeration listVertexesOriginal;
        GPJTriangle triangle;
        GPJPoint3D  vertexOriginal;

        if (m_lightOther != null &&
                m_shading == GOURAUD_SHADING)
        {
            for (listTriangles = m_listTriangles.elements(); listTriangles.hasMoreElements();)
            {
                triangle = (GPJTriangle)listTriangles.nextElement();
                triangle.m_plane.set((GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[0]),
                        (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[1]),
                        (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[2]));
            }

            for (listVertexesOriginal = m_listVertexesOriginal.elements(), index1 = 0; listVertexesOriginal.hasMoreElements(); index1 ++)
            {
                vertexOriginal = (GPJPoint3D)listVertexesOriginal.nextElement();
                vertexOriginal.m_normal.set(0.0f,
                        0.0f,
                        0.0f);

                numNormals = 0;
                for (listTriangles = m_listTriangles.elements(); listTriangles.hasMoreElements();)
                {
                    triangle = (GPJTriangle)listTriangles.nextElement();
                    for (a = 0; a < GPJTriangle.NUM_VERTEXES; a ++)
                        if (triangle.m_listNumVertexes[a] == index1)
                        {
                            numNormals ++;

                            GPJVector3D.addVectors(vertexOriginal.m_normal,
                                    triangle.m_plane.m_normal,
                                    vertexOriginal.m_normal);

                            break;
                        }
                }

                if (numNormals > 0)
                {
                    GPJVector3D.multiplyVectorScalar(vertexOriginal.m_normal,
                            1.0f / numNormals,
                            vertexOriginal.m_normal);
                    GPJVector3D.normalizeVector(vertexOriginal.m_normal,
                            vertexOriginal.m_normal);

                    if (m_lightOther.m_type == GPJLight.TYPE_DIRECTIONAL)
                        ratioColorLightOther = GPJVector3D.dotProductVectors(m_lightOther.m_vectorRot,
                                vertexOriginal.m_normal);
                    else if (m_lightOther.m_type == GPJLight.TYPE_POINT ||
                            m_lightOther.m_type == GPJLight.TYPE_SPOT)
                    {
                        GPJVector3D.subVectors(m_lightOther.m_pos,
                                vertexOriginal,
                                m_vectorTemp);
                        GPJVector3D.normalizeVector(m_vectorTemp,
                                m_vectorTemp);

                        ratioColorLightOther = GPJVector3D.dotProductVectors(m_vectorTemp,
                                vertexOriginal.m_normal);

                        if (m_lightOther.m_type == GPJLight.TYPE_SPOT)
                        {
                            if (ratioColorLightOther > 0.0f)
                                ratioColorLightOther *= GPJVector3D.dotProductVectors(m_lightOther.m_vectorRot,
                                    vertexOriginal.m_normal);
                            else
                                ratioColorLightOther = 0.0f;
                        }
                    }
                    else
                        ratioColorLightOther = 0.0f;
                    if (ratioColorLightOther < 0.0f)
                        ratioColorLightOther = 0.0f;

                    if (m_lightAmbient != null)
                        GPJColor.copyColor(m_lightAmbient.m_color,
                                m_colorTemp);
                    else
                        m_colorTemp.set(0,
                                0,
                                0);
                    GPJColor.interpolatesColors(m_colorTemp,
                            m_lightOther.m_color,
                            m_colorTemp2,
                            ratioColorLightOther);
                    GPJColor.multiplyColors(vertexOriginal.m_color,
                            m_colorTemp2,
                            vertexOriginal.m_colorProcessed);
                }
                else
                {
                    GPJColor.copyColor(vertexOriginal.m_color,
                            vertexOriginal.m_colorProcessed);
                }
            }
        }
        else
        {
            for (listVertexesOriginal = m_listVertexesOriginal.elements(); listVertexesOriginal.hasMoreElements();)
            {
                vertexOriginal = (GPJPoint3D)listVertexesOriginal.nextElement();

                if (m_lightAmbient != null      &&
                        m_lightOther == null    &&
                        m_shading == GOURAUD_SHADING)
                    GPJColor.multiplyColors(vertexOriginal.m_color,
                            m_lightAmbient.m_color,
                            vertexOriginal.m_colorProcessed);
                else
                    GPJColor.copyColor(vertexOriginal.m_color,
                            vertexOriginal.m_colorProcessed);
            }
        }
    }

    private void clipLines()
    {
        float       diffDistMinVertexZ;
        float       diffDistMaxVertexZ;
        float       diffVertexesZ;
        float       ratioDiffZ;
        float       distMinZ = m_camera.m_nearZ;
        float       distMaxZ = m_camera.m_farZ;
        Enumeration listLines;
        GPJLine     line;
        GPJPoint3D  vertexClipped;
        GPJPoint3D  vertex1;
        GPJPoint3D  vertex2;

        m_numVertexesClipped = 0;

        for (listLines = m_listLines.elements(); listLines.hasMoreElements();)
        {
            line = (GPJLine)listLines.nextElement();
            line.removeAllVertexesClipped();

            if (!line.isVisible())
                continue;

            vertex1 = line.m_listVertexesProcessed[0];
            vertex2 = line.m_listVertexesProcessed[1];

            if (vertex1.m_trasformTo2D &&
                    vertex2.m_trasformTo2D)
            {
                line.addVertexClipped(vertex1);
                line.addVertexClipped(vertex2);
            }
            else if (vertex1.m_trasformTo2D ||
                    vertex2.m_trasformTo2D)
            {
                diffVertexesZ = vertex2.m_z - vertex1.m_z;

                if (vertex1.m_z < distMinZ ||
                        vertex2.m_z < distMinZ)
                {
                    diffDistMinVertexZ  = distMinZ - vertex1.m_z;
                    ratioDiffZ          = diffDistMinVertexZ / diffVertexesZ;

                    if (diffVertexesZ < 0)
                    {
                        GPJColor.interpolatesColors(vertex1.m_color,
                                vertex2.m_color,
                                m_colorInterpolated,
                                ratioDiffZ);
                    }
                    else
                    {
                        GPJColor.interpolatesColors(vertex2.m_color,
                                vertex1.m_color,
                                m_colorInterpolated,
                                1.0f - ratioDiffZ);
                    }

                    m_colorInterpolated.calcRGB();

                    vertexClipped = m_listVertexesClipped[m_numVertexesClipped ++];
                    vertexClipped.set(vertex1.m_x + (vertex2.m_x - vertex1.m_x) * ratioDiffZ,
                            vertex1.m_y + (vertex2.m_y - vertex1.m_y) * ratioDiffZ,
                            distMinZ);
                    vertexClipped.m_trasformTo2D = true;

                    GPJColor.copyColor(m_colorInterpolated,
                            vertexClipped.m_color);

                    if (vertex1.m_z >= distMinZ &&
                            vertex2.m_z < distMinZ)
                    {
                        line.addVertexClipped(vertex1);
                        line.addVertexClipped(vertexClipped);
                    }
                    else
                    {
                        line.addVertexClipped(vertexClipped);
                        line.addVertexClipped(vertex2);
                    }
                }
                else
                {
                    diffDistMaxVertexZ  = vertex2.m_z - distMaxZ;
                    ratioDiffZ          = diffDistMaxVertexZ / diffVertexesZ;

                    if (diffVertexesZ >= 0)
                    {
                        GPJColor.interpolatesColors(vertex1.m_color,
                                vertex2.m_color,
                                m_colorInterpolated,
                                1.0f - ratioDiffZ);
                    }
                    else
                    {
                        GPJColor.interpolatesColors(vertex2.m_color,
                                vertex1.m_color,
                                m_colorInterpolated,
                                ratioDiffZ);
                    }

                    m_colorInterpolated.calcRGB();

                    vertexClipped = m_listVertexesClipped[m_numVertexesClipped ++];
                    vertexClipped.set(vertex1.m_x + (vertex2.m_x - vertex1.m_x) * (1.0f - ratioDiffZ),
                            vertex1.m_y + (vertex2.m_y - vertex1.m_y) * (1.0f - ratioDiffZ),
                            distMaxZ);
                    vertexClipped.m_trasformTo2D = true;

                    GPJColor.copyColor(m_colorInterpolated,
                            vertexClipped.m_color);

                    if (vertex1.m_z <= distMaxZ &&
                            vertex2.m_z > distMaxZ)
                    {
                        line.addVertexClipped(vertex1);
                        line.addVertexClipped(vertexClipped);
                    }
                    else
                    {
                        line.addVertexClipped(vertexClipped);
                        line.addVertexClipped(vertex2);
                    }
                }
            }
        }
    }
    
    private void clipTriangles()
    {
        boolean     first_vertexes;
        int         i;
        int         numVertex1;
        int         numVertex2;
        float       diffDistMinVertexZ;
        float       diffDistMaxVertexZ;
        float       diffVertexesZ;
        float       ratioDiffZ;
        float       distMinZ = m_camera.m_nearZ;
        float       distMaxZ = m_camera.m_farZ;
        Enumeration listTriangles;
        GPJTriangle triangle;
        GPJPoint3D  vertexClipped;
        GPJPoint3D  vertex1;
        GPJPoint3D  vertex2;
        
        for (listTriangles = m_listTriangles.elements(); listTriangles.hasMoreElements();)
        {
            triangle = (GPJTriangle)listTriangles.nextElement();
            triangle.removeAllVertexesClipped();

            if (!triangle.isVisible())
                continue;

            first_vertexes = true;

            for (i = 0; i < GPJTriangle.NUM_VERTEXES; i ++)
            {
                numVertex1 = i;
                numVertex2 = numVertex1 + 1;
                if (numVertex2 == GPJTriangle.NUM_VERTEXES)
                    numVertex2 = 0;

                vertex1 = triangle.m_listVertexesProcessed[numVertex1];
                vertex2 = triangle.m_listVertexesProcessed[numVertex2];
                
                if (vertex1.m_trasformTo2D &&
                        vertex2.m_trasformTo2D)
                {
                    if (first_vertexes)
                    {
                        triangle.addVertexClipped(vertex1);
                        triangle.addVertexClipped(vertex2);
                    }
                    else
                        triangle.addVertexClipped(vertex2);

                    first_vertexes = false;
                } 
                else if (vertex1.m_trasformTo2D ||
                        vertex2.m_trasformTo2D)
                {
                    diffVertexesZ = vertex2.m_z - vertex1.m_z;

                    if (vertex1.m_z < distMinZ ||
                            vertex2.m_z < distMinZ)
                    {
                        diffDistMinVertexZ  = distMinZ - vertex1.m_z;
                        ratioDiffZ          = diffDistMinVertexZ / diffVertexesZ;

                        if (diffVertexesZ < 0)
                        {
                            GPJColor.interpolatesColors(vertex1.m_color,
                                    vertex2.m_color,
                                    m_colorInterpolated,
                                    ratioDiffZ);
                            
                            GPJVector2D.interpolatesVectors(vertex1.m_uv, 
                                    vertex2.m_uv,
                                    m_uvInterpolated, 
                                    ratioDiffZ);
                        }
                        else
                        {
                            GPJColor.interpolatesColors(vertex2.m_color,
                                    vertex1.m_color,
                                    m_colorInterpolated,
                                    1.0f - ratioDiffZ);
                            
                            GPJVector2D.interpolatesVectors(vertex2.m_uv, 
                                    vertex1.m_uv,
                                    m_uvInterpolated, 
                                    1.0f - ratioDiffZ);
                        }

                        m_colorInterpolated.calcRGB();

                        vertexClipped = m_listVertexesClipped[m_numVertexesClipped ++];
                        vertexClipped.set(vertex1.m_x + (vertex2.m_x - vertex1.m_x) * ratioDiffZ,
                                vertex1.m_y + (vertex2.m_y - vertex1.m_y) * ratioDiffZ,
                                distMinZ);
                        vertexClipped.m_trasformTo2D = true;

                        GPJColor.copyColor(m_colorInterpolated,
                                vertexClipped.m_color);

                        GPJVector2D.copyVector(m_uvInterpolated,
                                vertexClipped.m_uv);

                        if (vertex1.m_z >= distMinZ &&
                                vertex2.m_z < distMinZ)
                        {
                            if (first_vertexes)
                            {
                                triangle.addVertexClipped(vertex1);
                                triangle.addVertexClipped(vertexClipped);
                            }
                            else
                            {
                                triangle.addVertexClipped(vertexClipped);
                                if (numVertex2 == 0)
                                    triangle.addFirstVertexClipped();
                            }
                        }
                        else
                        {
                            triangle.addVertexClipped(vertexClipped);
                            triangle.addVertexClipped(vertex2);
                        }

                        first_vertexes = false;
                    }
                    else
                    {
                        diffDistMaxVertexZ  = vertex2.m_z - distMaxZ;
                        ratioDiffZ          = diffDistMaxVertexZ / diffVertexesZ;

                        if (diffVertexesZ >= 0)
                        {
                            GPJColor.interpolatesColors(vertex1.m_color,
                                    vertex2.m_color,
                                    m_colorInterpolated,
                                    1.0f - ratioDiffZ);
                            
                            GPJVector2D.interpolatesVectors(vertex1.m_uv, 
                                    vertex2.m_uv, 
                                    m_uvInterpolated, 
                                    1.0f - ratioDiffZ);
                        }
                        else
                        {
                            GPJColor.interpolatesColors(vertex2.m_color,
                                    vertex1.m_color,
                                    m_colorInterpolated,
                                    ratioDiffZ);
                            
                            GPJVector2D.interpolatesVectors(vertex2.m_uv, 
                                    vertex1.m_uv, 
                                    m_uvInterpolated, 
                                    ratioDiffZ);
                        }

                        m_colorInterpolated.calcRGB();
                      
                        vertexClipped = m_listVertexesClipped[m_numVertexesClipped ++];
                        vertexClipped.set(vertex1.m_x + (vertex2.m_x - vertex1.m_x) * (1.0f - ratioDiffZ),
                                vertex1.m_y + (vertex2.m_y - vertex1.m_y) * (1.0f - ratioDiffZ),
                                distMaxZ);
                        vertexClipped.m_trasformTo2D = true;

                        GPJColor.copyColor(m_colorInterpolated,
                                vertexClipped.m_color);

                        GPJVector2D.copyVector(m_uvInterpolated,
                                vertexClipped.m_uv);

                        if (vertex1.m_z <= distMaxZ &&
                                vertex2.m_z > distMaxZ)
                        {
                            if (first_vertexes)
                            {
                                triangle.addVertexClipped(vertex1);
                                triangle.addVertexClipped(vertexClipped);
                            }
                            else
                            {
                                triangle.addVertexClipped(vertexClipped);
                                if (numVertex2 == 0)
                                    triangle.addFirstVertexClipped();
                            }
                        }
                        else
                        {
                            triangle.addVertexClipped(vertexClipped);
                            triangle.addVertexClipped(vertex2);
                        }

                        first_vertexes = false;
                    }
                }
            }
        }
    }
    
    private void trasformVertexes3Dto2D()
    {
        int         i;
        float       z;
        Enumeration listVertexes;
        GPJPoint3D  vertex;
        
        for (listVertexes = m_listVertexesProcessed.elements(); listVertexes.hasMoreElements();)
        {
            vertex = (GPJPoint3D)listVertexes.nextElement();
            vertex.trasformTo2D(m_camera.m_screenCenterX,
                    m_camera.m_screenCenterY,
                    m_camera.m_screenDistX,
                    m_camera.m_screenDistY);
        }        
        for (i = 0; i < m_numVertexesClipped; i ++)
        {
            vertex = m_listVertexesClipped[i];
            vertex.trasformTo2D(m_camera.m_screenCenterX,
                    m_camera.m_screenCenterY,
                    m_camera.m_screenDistX,
                    m_camera.m_screenDistY);
        }
    }

    private void trasformPoints3Dto2D()
    {
        Enumeration listPoints;
        GPJPoint3D  point;

        for (listPoints = m_listPointsProcessed.elements(); listPoints.hasMoreElements();)
        {
            point = (GPJPoint3D)listPoints.nextElement();
            point.trasformTo2D(m_camera.m_screenCenterX,
                    m_camera.m_screenCenterY,
                    m_camera.m_screenDistX,
                    m_camera.m_screenDistY);
        }
    }
    
    private void draw()
    {
        Enumeration listPoints;
        Enumeration listLines;
        Enumeration listTriangles;
        GPJPoint3D  point;
        GPJLine     line;
        GPJTriangle triangle;

        for (listPoints = m_listPointsProcessed.elements(); listPoints.hasMoreElements();)
        {
            point = (GPJPoint3D)listPoints.nextElement();
            if (point.isVisible())
                drawPoint(point);
        }

        for (listTriangles = m_listTriangles.elements(); listTriangles.hasMoreElements();)
        {
            triangle = (GPJTriangle)listTriangles.nextElement();
            if (triangle.m_numVertexesClipped > 0)
                drawTriangle(triangle);
        }

        for (listLines = m_listLines.elements(); listLines.hasMoreElements();)
        {
            line = (GPJLine)listLines.nextElement();
            if (line.m_numVertexesClipped > 0)
                drawLine(line);
        }
    }

    private void drawLine(GPJLine line)
    {
        boolean steep;
        int     l;
        int     deltaX;
        int     deltaY;
        int     error;
        int     x;
        int     y;
        int     x0;
        int     y0;
        int     x1;
        int     y1;
        int     stepY;
        int     firstIndex;
        int     secondIndex;
        int     temp;
        float   a;
        float   b;
        float   c;
        float   d;
        float   e;
        float   ax;
        float   ay;
        float   bx;
        float   by;
        float   z;
        float   rate;
        float   numer;
        float   denom;

        firstIndex  = 0;
        secondIndex = 1;

        x0 = line.m_listVertexesClipped[firstIndex].m_xOnScreen;
        y0 = line.m_listVertexesClipped[firstIndex].m_yOnScreen;
        x1 = line.m_listVertexesClipped[secondIndex].m_xOnScreen;
        y1 = line.m_listVertexesClipped[secondIndex].m_yOnScreen;

        if (Math.abs(x0 - x1) <= 1 &&
                Math.abs(y0 - y1) <= 1)
        {
            if (x0 >= 0                         &&
                    y0 >= 0                     &&
                    x0 < m_camera.m_screenWidth &&
                    y0 < m_camera.m_screenHeight)
            {
                drawLinePoint(line,
                        x0 + y0 * m_camera.m_screenWidth,
                        0.0f,
                        line.m_listVertexesClipped[firstIndex].m_z,
                        firstIndex,
                        secondIndex);
            }

            if ((x0 != x1                       ||
                    y0 != y1)                   &&
                    x1 >= 0                     &&
                    y1 >= 0                     &&
                    x1 < m_camera.m_screenWidth &&
                    y1 < m_camera.m_screenHeight)
            {
                drawLinePoint(line,
                        x1 + y1 * m_camera.m_screenWidth,
                        1.0f,
                        line.m_listVertexesClipped[secondIndex].m_z,
                        firstIndex,
                        secondIndex);
            }

            return;
        }
        
        if (Math.abs(y1 - y0) > Math.abs(x1 - x0))
            steep = true;
        else
            steep = false;
        if (steep)
        {
            l   = y0;
            y0  = x0;
            x0  = l;
            
            l   = y1;
            y1  = x1;
            x1  = l;
        }
        if (x1 < x0)
        {
            l   = x1;
            x1  = x0;
            x0  = l;

            l   = y1;
            y1  = y0;
            y0  = l;

            temp        = firstIndex;
            firstIndex  = secondIndex;
            secondIndex = temp;
        }

        deltaX  = x1 - x0;
        deltaY  = Math.abs(y1 - y0);
        error   = deltaX >> 1;

        if (y0 < y1)
            stepY = 1;
        else
            stepY = -1;
        
        GPJVector3D.subVectors(line.m_listVertexesClipped[secondIndex], 
                line.m_listVertexesClipped[firstIndex], 
                m_renderA);                                 // V1

        ax = m_camera.m_nearZ / m_camera.m_screenDistX;
        ay = m_camera.m_nearZ / m_camera.m_screenDistY;
        bx = m_camera.m_farZ / m_camera.m_screenDistX;
        by = m_camera.m_farZ / m_camera.m_screenDistY;

        y = y0;
        for (x = x0; x <= x1; x ++)
        {
            if (steep)
            {
                if (y >= 0                          &&
                        x >= 0                      &&
                        y < m_camera.m_screenWidth  &&
                        x < m_camera.m_screenHeight)
                {
                    m_renderB.set((y - m_camera.m_screenCenterX) * ax,
                            (x - m_camera.m_screenCenterY) * ay,
                            m_camera.m_nearZ);              // P2

                    m_renderC.set((y - m_camera.m_screenCenterX) * bx,
                            (x - m_camera.m_screenCenterY) * by,
                            m_camera.m_farZ);

                    GPJVector3D.subVectors(m_renderC,
                            m_renderB,
                            m_renderC);                     // V2
                    
                    GPJVector3D.subVectors(line.m_listVertexesClipped[firstIndex], 
                            m_renderB, 
                            m_renderD);                     // P1 - P2

                    a = GPJVector3D.dotProductVectors(m_renderD,
                            m_renderC);
                    b = GPJVector3D.dotProductVectors(m_renderC,
                            m_renderA);
                    c = GPJVector3D.dotProductVectors(m_renderD,
                            m_renderA);
                    d = m_renderC.getMagnitude2();
                    e = m_renderA.getMagnitude2();

                    denom = d * e - b * b;
                    if (Math.abs(denom) <= TOLLERANCE)
                        continue;

                    numer = a * b - c * d;

                    rate = numer / denom;

                    if (rate < 0.0f ||
                            rate > 1.0f)
                        continue;

                    drawLinePoint(line,
                            y + x * m_camera.m_screenWidth,
                            rate,
                            line.m_listVertexesClipped[firstIndex].m_z + rate * m_renderA.m_z,
                            firstIndex,
                            secondIndex);
                }
            }
            else
            {
                if (x >= 0                          &&
                        y >= 0                      &&
                        x < m_camera.m_screenWidth  &&
                        y < m_camera.m_screenHeight)
                {
                    m_renderB.set((x - m_camera.m_screenCenterX) * ax,
                            (y - m_camera.m_screenCenterY) * ay,
                            m_camera.m_nearZ);              // P2

                    m_renderC.set((x - m_camera.m_screenCenterX) * bx,
                            (y - m_camera.m_screenCenterY) * by,
                            m_camera.m_farZ);

                    GPJVector3D.subVectors(m_renderC,
                            m_renderB,
                            m_renderC);                     // V2

                    GPJVector3D.subVectors(line.m_listVertexesClipped[firstIndex],
                            m_renderB,
                            m_renderD);                     // P1 - P2

                    a = GPJVector3D.dotProductVectors(m_renderD,
                            m_renderC);
                    b = GPJVector3D.dotProductVectors(m_renderC,
                            m_renderA);
                    c = GPJVector3D.dotProductVectors(m_renderD,
                            m_renderA);
                    d = m_renderC.getMagnitude2();
                    e = m_renderA.getMagnitude2();

                    denom = d * e - b * b;
                    if (Math.abs(denom) <= TOLLERANCE)
                        continue;

                    numer = a * b - c * d;

                    rate = numer / denom;

                    if (rate < 0.0f ||
                            rate > 1.0f)
                        continue;

                    drawLinePoint(line,
                            x + y * m_camera.m_screenWidth,
                            rate,
                            line.m_listVertexesClipped[firstIndex].m_z + rate * m_renderA.m_z,
                            firstIndex,
                            secondIndex);
                }
            }
            error -= deltaY;
            if (error < 0)
            {
                y       += stepY;
                error   += deltaX;
            }
        }
    }

    private void drawLinePoint(GPJLine  line,
                               int      offset,
                               float    rate,
                               float    z,
                               int      firstIndex,
                               int      secondIndex)
    {
        if (!m_camera.setValueZBuffer(offset, z, true))
            return;

        GPJColor.interpolatesColors(line.m_listVertexesClipped[firstIndex].m_color,
                line.m_listVertexesClipped[secondIndex].m_color,
                m_colorTemp,
                rate);
        m_colorTemp.calcRGB();

        m_camera.m_screenBuffer[offset] = m_colorTemp.m_RGB;
    }

    private void drawTriangle(GPJTriangle triangle)
    {
        int         i;
        int         j;
        int         l;
        int         x;
        int         y;
        int         x0;
        int         y0;
        int         x1;
        int         y1;
        int         x2;
        int         y2;
        int         xx1;
        int         xx2;
        int         firstIndex;
        int         secondIndex;
        int         thirdIndex;
        int         offset1;
        int         offset2;
        int[]       deltaX;
        int[]       deltaY;
        float       a;
        float       a2;
        float       b;
        float       b2;
        float       c;
        float       c2;
        float       startX;
        float       endX;
        float       xf;
        float       yf;
        float       ratioColorLightOther;
        float[]     d;
        GPJPoint3D  vertexClipped;
        GPJPoint3D  oldVertexClipped = null;

        deltaX  = new int[GPJTriangle.NUM_VERTEXES];
        deltaY  = new int[GPJTriangle.NUM_VERTEXES];
        d       = new float[GPJTriangle.NUM_VERTEXES];

        if (m_shading == FLAT_SHADING)
        {
            if (m_lightOther != null)
            {
                triangle.m_plane.set((GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[0]),
                        (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[1]),
                        (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[2]));

                if (m_lightOther.m_type == GPJLight.TYPE_DIRECTIONAL)
                    ratioColorLightOther = GPJVector3D.dotProductVectors(m_lightOther.m_vectorRot,
                            triangle.m_plane.m_normal);
                else if (m_lightOther.m_type == GPJLight.TYPE_POINT ||
                        m_lightOther.m_type == GPJLight.TYPE_SPOT)
                {
                    GPJVector3D.copyVector((GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[0]),
                            m_vectorTemp);
                    GPJVector3D.addVectors(m_vectorTemp,
                            (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[1]),
                            m_vectorTemp);
                    GPJVector3D.addVectors(m_vectorTemp,
                            (GPJVector3D)m_listVertexesOriginal.elementAt(triangle.m_listNumVertexes[2]),
                            m_vectorTemp);
                    GPJVector3D.multiplyVectorScalar(m_vectorTemp,
                            1 / 3.0f,
                            m_vectorTemp);

                    GPJVector3D.subVectors(m_lightOther.m_pos,
                            m_vectorTemp,
                            m_vectorTemp);
                    GPJVector3D.normalizeVector(m_vectorTemp,
                            m_vectorTemp);

                    ratioColorLightOther = GPJVector3D.dotProductVectors(m_vectorTemp,
                            triangle.m_plane.m_normal);

                    if (m_lightOther.m_type == GPJLight.TYPE_SPOT)
                    {
                        if (ratioColorLightOther > 0.0f)
                            ratioColorLightOther *= GPJVector3D.dotProductVectors(m_lightOther.m_vectorRot,
                                triangle.m_plane.m_normal);
                        else
                            ratioColorLightOther = 0.0f;
                    }
                }
                else
                    ratioColorLightOther = 0.0f;
                if (ratioColorLightOther < 0.0f)
                    ratioColorLightOther = 0.0f;
            }
            else
                ratioColorLightOther = 0.0f;

            for (i = 0; i < triangle.m_numVertexesClipped - 1; i ++)
            {
                vertexClipped = triangle.m_listVertexesClipped[i];
                if (vertexClipped == oldVertexClipped)
                    continue;
                else
                    oldVertexClipped = vertexClipped;

                if (m_lightOther != null)
                {
                    if (m_lightAmbient != null)
                        GPJColor.copyColor(m_lightAmbient.m_color,
                                m_colorTemp);
                    else
                        m_colorTemp.set(0,
                                0,
                                0);
                    GPJColor.interpolatesColors(m_colorTemp,
                            m_lightOther.m_color,
                            m_colorTemp2,
                            ratioColorLightOther);
                    GPJColor.multiplyColors(vertexClipped.m_color,
                            m_colorTemp2,
                            vertexClipped.m_colorProcessed);
                }
                else if (m_lightAmbient != null)
                {
                    GPJColor.multiplyColors(vertexClipped.m_color,
                            m_lightAmbient.m_color,
                            vertexClipped.m_colorProcessed);
                }
                else
                    GPJColor.copyColor(vertexClipped.m_color,
                        vertexClipped.m_colorProcessed);
            }
        }

        for (i = 0; i <= triangle.m_numVertexesClipped - GPJTriangle.NUM_VERTEXES; i += 2)
        {
            firstIndex  = i;
            secondIndex = firstIndex + 1;
            thirdIndex  = secondIndex + 1;

            GPJVector3D.copyVector(triangle.m_listVertexesClipped[firstIndex],
                    m_renderOldP);

            GPJVector3D.subVectors(triangle.m_listVertexesClipped[secondIndex],
                    m_renderOldP,
                    m_renderOldM);

            GPJVector3D.subVectors(triangle.m_listVertexesClipped[thirdIndex],
                    m_renderOldP,
                    m_renderOldN);

            m_renderNewP.set(m_renderOldP.m_x * m_camera.m_screenDistX,
                    m_renderOldP.m_y * m_camera.m_screenDistY,
                    m_renderOldP.m_z);

            m_renderNewM.set(m_renderOldM.m_x * m_camera.m_screenDistX,
                    m_renderOldM.m_y * m_camera.m_screenDistY,
                    m_renderOldM.m_z);

            m_renderNewN.set(m_renderOldN.m_x * m_camera.m_screenDistX,
                    m_renderOldN.m_y * m_camera.m_screenDistY,
                    m_renderOldN.m_z);

            GPJVector3D.crossProductVectors(m_renderNewN,
                    m_renderNewP,
                    m_renderA);

            GPJVector3D.crossProductVectors(m_renderNewP,
                    m_renderNewM,
                    m_renderB);

            GPJVector3D.crossProductVectors(m_renderNewM,
                    m_renderNewN,
                    m_renderC);

            x0 = triangle.m_listVertexesClipped[firstIndex].m_xOnScreen;
            y0 = triangle.m_listVertexesClipped[firstIndex].m_yOnScreen;
            x1 = triangle.m_listVertexesClipped[secondIndex].m_xOnScreen;
            y1 = triangle.m_listVertexesClipped[secondIndex].m_yOnScreen;
            x2 = triangle.m_listVertexesClipped[thirdIndex].m_xOnScreen;
            y2 = triangle.m_listVertexesClipped[thirdIndex].m_yOnScreen;

            if (y1 < y0)
            {
                l   = y1;
                y1  = y0;
                y0  = l;

                l   = x1;
                x1  = x0;
                x0  = l;
            }
            if (y2 < y0)
            {
                l   = y2;
                y2  = y0;
                y0  = l;

                l   = x2;
                x2  = x0;
                x0  = l;
            }
            if (y2 < y1)
            {
                l   = y2;
                y2  = y1;
                y1  = l;

                l   = x2;
                x2  = x1;
                x1  = l;
            }

            deltaX[0] = x1 - x0;
            deltaY[0] = y1 - y0;
            deltaX[1] = x2 - x1;
            deltaY[1] = y2 - y1;
            deltaX[2] = x2 - x0;
            deltaY[2] = y2 - y0;

            for (j = 0; j < GPJTriangle.NUM_VERTEXES; j ++)
            {
                if (deltaY[j] != 0)
                    d[j] = (float)deltaX[j] / (float)deltaY[j];
                else
                    d[j] = 0.0f;
            }

            startX  = x0;
            endX    = x0;
            offset1 = y0 * m_camera.m_screenWidth;
            yf      = y0 - m_camera.m_screenCenterY + 0.5f;
            a       = m_renderA.m_z + yf * m_renderA.m_y;
            b       = m_renderB.m_z + yf * m_renderB.m_y;
            c       = m_renderC.m_z + yf * m_renderC.m_y;
            for (y = y0; y < y1; y ++, startX += d[0], endX += d[2], offset1 += m_camera.m_screenWidth, yf += 1.0f, a += m_renderA.m_y, b += m_renderB.m_y, c += m_renderC.m_y)
            {
                if (y < 0)
                    continue;
                else if (y >= m_camera.m_screenHeight)
                    break;

                if (startX < endX)
                {
                    xx1 = (int)startX;
                    xx2 = (int)endX;
                }
                else
                {
                    xx1 = (int)endX;
                    xx2 = (int)startX;
                }
                offset2 = offset1 + xx1;
                xf      = xx1 - m_camera.m_screenCenterX + 0.5f;
                a2      = a + xf * m_renderA.m_x;
                b2      = b + xf * m_renderB.m_x;
                c2      = c + xf * m_renderC.m_x;
                for (x = xx1; x <= xx2; x ++, offset2 ++, xf += 1.0f, a2 += m_renderA.m_x, b2 += m_renderB.m_x, c2 += m_renderC.m_x)
                {
                    if (x < 0 ||
                            c2 == 0.0f)
                        continue;
                    else if (x >= m_camera.m_screenWidth)
                        break;

                    drawTrianglePoint(triangle,
                            offset2,
                            firstIndex,
                            secondIndex,
                            thirdIndex,
                            a2,
                            b2,
                            c2);
                }
            }

            startX  = x1;
            endX    = x0 + (y1 - y0) * d[2];
            offset1 = y1 * m_camera.m_screenWidth;
            yf      = y1 - m_camera.m_screenCenterY + 0.5f;
            a       = m_renderA.m_z + yf * m_renderA.m_y;
            b       = m_renderB.m_z + yf * m_renderB.m_y;
            c       = m_renderC.m_z + yf * m_renderC.m_y;
            for (y = y1; y <= y2; y ++, startX += d[1], endX += d[2], offset1 += m_camera.m_screenWidth, yf += 1.0f, a += m_renderA.m_y, b += m_renderB.m_y, c += m_renderC.m_y)
            {
                if (y < 0)
                    continue;
                else if (y >= m_camera.m_screenHeight)
                    break;

                if (startX < endX)
                {
                    xx1 = (int)startX;
                    xx2 = (int)endX;
                }
                else
                {
                    xx1 = (int)endX;
                    xx2 = (int)startX;
                }
                offset2 = offset1 + xx1;
                xf      = xx1 - m_camera.m_screenCenterX + 0.5f;
                a2      = a + xf * m_renderA.m_x;
                b2      = b + xf * m_renderB.m_x;
                c2      = c + xf * m_renderC.m_x;
                for (x = xx1; x <= xx2; x ++, offset2 ++, xf += 1.0f, a2 += m_renderA.m_x, b2 += m_renderB.m_x, c2 += m_renderC.m_x)
                {
                    if (x < 0 ||
                            c2 == 0.0f)
                        continue;
                    else if (x >= m_camera.m_screenWidth)
                        break;

                    drawTrianglePoint(triangle,
                            offset2,
                            firstIndex,
                            secondIndex,
                            thirdIndex,
                            a2,
                            b2,
                            c2);
                }
            }
        }
    }

    private void drawTrianglePoint(GPJTriangle  triangle,
                                   int          offset,
                                   int          firstIndex,
                                   int          secondIndex,
                                   int          thirdIndex,
                                   float        a,
                                   float        b,
                                   float        c)
    {
        int     xx = 0;
        int     yy = 0;
        float   u;
        float   v;
        float   newU;
        float   newV;
        float   d;
        
        u = a / c;
        v = b / c;

        if (u < 0.0f)
            u = 0.0f;
        if (v < 0.0f)
            v = 0.0f;

        if (u > 1.0f)
            u = 1.0f;
        if (v > 1.0f)
            v = 1.0f;

        /*m_pointTemp.set(oldP.m_x + u * oldM.m_x + v * oldN.m_x,
                oldP.m_y + u * oldM.m_y + v * oldN.m_y,
                oldP.m_z + u * oldM.m_z + v * oldN.m_z);*/

        if (!m_camera.setValueZBuffer(offset, m_renderOldP.m_z + u * m_renderOldM.m_z + v * m_renderOldN.m_z/*m_pointTemp.m_z*/, true))
            return;

        d = 1.0f - u - v;
        if (d < 0.0f)
        {
            d /= 2.0f;
            u += d;
            v += d;
            d = 0.0f;
        }

        if (m_shading == FLAT_SHADING)
        {
            m_colorTemp.m_red   = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_colorProcessed.m_red + v * triangle.m_listVertexesClipped[thirdIndex].m_colorProcessed.m_red + d * triangle.m_listVertexesClipped[firstIndex].m_colorProcessed.m_red);
            m_colorTemp.m_green = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_colorProcessed.m_green + v * triangle.m_listVertexesClipped[thirdIndex].m_colorProcessed.m_green + d * triangle.m_listVertexesClipped[firstIndex].m_colorProcessed.m_green);
            m_colorTemp.m_blue  = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_colorProcessed.m_blue + v * triangle.m_listVertexesClipped[thirdIndex].m_colorProcessed.m_blue + d * triangle.m_listVertexesClipped[firstIndex].m_colorProcessed.m_blue);
        }
        else
        {
            m_colorTemp.m_red   = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_color.m_red + v * triangle.m_listVertexesClipped[thirdIndex].m_color.m_red + d * triangle.m_listVertexesClipped[firstIndex].m_color.m_red);
            m_colorTemp.m_green = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_color.m_green + v * triangle.m_listVertexesClipped[thirdIndex].m_color.m_green + d * triangle.m_listVertexesClipped[firstIndex].m_color.m_green);
            m_colorTemp.m_blue  = (int)(u * triangle.m_listVertexesClipped[secondIndex].m_color.m_blue + v * triangle.m_listVertexesClipped[thirdIndex].m_color.m_blue + d * triangle.m_listVertexesClipped[firstIndex].m_color.m_blue);
        }

        if (triangle.m_texture != null)
        {
            newU  = u * triangle.m_listVertexesClipped[secondIndex].m_uv.m_x + v * triangle.m_listVertexesClipped[thirdIndex].m_uv.m_x + d * triangle.m_listVertexesClipped[firstIndex].m_uv.m_x;
            newV  = u * triangle.m_listVertexesClipped[secondIndex].m_uv.m_y + v * triangle.m_listVertexesClipped[thirdIndex].m_uv.m_y + d * triangle.m_listVertexesClipped[firstIndex].m_uv.m_y;

            xx = (int)(newU * (float)(triangle.m_texture.getWidth() - 1));
            yy = (int)(newV * (float)(triangle.m_texture.getHeight() - 1));

            m_colorTemp2.m_RGB = triangle.m_texture.m_textureBuffer[xx + yy * triangle.m_texture.getWidth()] & 0x00ffffff;
            m_colorTemp2.calcRedGreenBlue();
            GPJColor.multiplyColors(m_colorTemp,
                    m_colorTemp2,
                    m_colorTemp);
        }

        if (triangle.m_texture != null &&
                triangle.m_texture.m_alphaBuffer != null)
        {
            m_colorTemp2.m_RGB = m_camera.m_screenBuffer[offset];
            m_colorTemp2.calcRedGreenBlue();
            GPJColor.interpolatesColors(m_colorTemp2,
                    m_colorTemp,
                    m_colorTemp,
                    triangle.m_texture.m_alphaBuffer[xx + yy * triangle.m_texture.getWidth()]);
        }

        m_colorTemp.calcRGB();

        m_camera.m_screenBuffer[offset] = m_colorTemp.m_RGB;
    }
 
    private void drawPoint(GPJPoint3D point)
    {
        int x = point.m_xOnScreen;
        int y = point.m_yOnScreen;
        int offset;

        if (x < 0                           ||
                y < 0                       ||
                x >= m_camera.m_screenWidth ||
                y >= m_camera.m_screenHeight)
            return;
        
        offset = x + y * m_camera.m_screenWidth;

        if (m_camera.setValueZBuffer(offset, point.m_z, true))
            m_camera.m_screenBuffer[offset] = point.m_color.m_RGB;
    }

}
